#include "pqOptUMesh.h"
#include "pqPropertiesPanel.h"
#include "pvOptFuncs.cxx"
#include "pvOptFuncs.h"
#include "stlFuncs.cxx"
#include "math.h"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstring>
#include <sstream>
#include <string>
#include <vector>

#include <QDebug>
#include <QVBoxLayout>
#include <QGridLayout>
#include <QPushButton>
#include <QString>
#include <QStringList>
#include <QObject>
#include <QFile>
#include <QFileInfo>
#include <QFileDialog>
#include <QWidget>
#include <QLabel>
#include <QLineEdit>
#include <QMessageBox>
#include <QFuture>
#include <QtConcurrent/QtConcurrent>

#include <vtksys/SystemTools.hxx>
#include <vtkSmartPointer.h>
#include <vtkCubeSource.h>
#include <vtkActor.h>
#include <vtkPolyData.h>
#include <vtkPolyDataMapper.h>
#include <vtkProperty.h>
#include <vtkRenderWindow.h>
#include <vtkRenderWindowInteractor.h>
#include <vtkRenderer.h>
#include <vtkBorderWidget.h>
#include <vtkBorderRepresentation.h>
#include <vtkFeatureEdges.h>
#include <vtkAssembly.h>
#include <vtkNew.h>
#include <vtkSTLReader.h>
#include <vtkNamedColors.h>
      
//-----------------------------------------------------------------------------

pqOptUMesh::pqOptUMesh(
  vtkSMProxy* smproxy, vtkSMProperty* smproperty, QWidget* parentObject)
  : Superclass(smproxy, parentObject)
{
  this->setShowLabel(false);

  QVBoxLayout* masterLayout = new QVBoxLayout(this);

  QGridLayout* gridLayout = new QGridLayout(this);
  gridLayout->setMargin(pqPropertiesPanel::suggestedMargin());
  gridLayout->setHorizontalSpacing(pqPropertiesPanel::suggestedHorizontalSpacing());
  gridLayout->setVerticalSpacing(pqPropertiesPanel::suggestedVerticalSpacing());
  gridLayout->setColumnStretch(1, 0);
  masterLayout->addLayout(gridLayout,0);

  QLabel* spacer = new QLabel(this);

  //link case file button
  QPushButton* caseFileHandlerButton = new QPushButton(this);
  caseFileHandlerButton->setText(tr("Select Output Folder"));
  connect(caseFileHandlerButton, SIGNAL(clicked()), SLOT(caseFileHandler()));
  gridLayout->addWidget(caseFileHandlerButton);

  //About button
  QPushButton* aboutButton = new QPushButton(this);
  aboutButton->setText(tr("About pvOptUMesh"));
  connect(aboutButton, SIGNAL(clicked()), SLOT(aboutMessage()));
  gridLayout->addWidget(aboutButton);

  //Select tri-surface mesh file
  QPushButton* selectSurfaceButton = new QPushButton(this);
  selectSurfaceButton->setText(tr("Select Tri-Surface Mesh Files"));
  connect(selectSurfaceButton, SIGNAL(clicked()), SLOT(selectSurface()));
  gridLayout->addWidget(selectSurfaceButton);

  // button to show block visualization
  QPushButton* blockVizButton = new QPushButton(this);
  blockVizButton->setText(tr("Visualize Background Mesh"));
  connect(blockVizButton, SIGNAL(clicked()), SLOT(visualizeBoundBox()));
  gridLayout->addWidget(blockVizButton);

  //block mesh setup
  QLabel* blockMeshLabel = new QLabel(tr("Block Mesh Parameters"), this);
  gridLayout->addWidget(blockMeshLabel);
  gridLayout->addWidget(spacer);

  //blockmesh points
  QGridLayout* secondGridLayout = new QGridLayout(this);
  secondGridLayout->setMargin(pqPropertiesPanel::suggestedMargin());
  secondGridLayout->setHorizontalSpacing(pqPropertiesPanel::suggestedHorizontalSpacing());
  secondGridLayout->setVerticalSpacing(pqPropertiesPanel::suggestedVerticalSpacing());
  secondGridLayout->setColumnStretch(2, 0);
  masterLayout->addLayout(secondGridLayout, 0);

  QLabel* blockMeshPointsLabel = new QLabel(tr("Block Mesh Points"), this);
  secondGridLayout->addWidget(blockMeshPointsLabel);
  QLabel* minLabel = new QLabel(tr("Min"), this);
  minLabel->setAlignment(Qt::AlignHCenter);
  secondGridLayout->addWidget(minLabel);
  QLabel* maxLabel = new QLabel(tr("Max"), this);
  maxLabel->setAlignment(Qt::AlignHCenter);
  secondGridLayout->addWidget(maxLabel);

  QLabel* blockMeshPointsLabelX = new QLabel(tr("X Axis"), this);
  secondGridLayout->addWidget(blockMeshPointsLabelX);
  blockMeshXMin = new QLineEdit(this);
  blockMeshXMin->setText("0");
  secondGridLayout->addWidget(blockMeshXMin);
  blockMeshXMax = new QLineEdit(this);
  blockMeshXMax->setText("1");
  secondGridLayout->addWidget(blockMeshXMax);

  QLabel* blockMeshPointsLabelY = new QLabel(tr("Y Axis"), this);
  secondGridLayout->addWidget(blockMeshPointsLabelY);
  blockMeshYMin = new QLineEdit(this);
  blockMeshYMin->setText("0");
  secondGridLayout->addWidget(blockMeshYMin);
  blockMeshYMax = new QLineEdit(this);
  blockMeshYMax->setText("1");
  secondGridLayout->addWidget(blockMeshYMax);
  
  QLabel* blockMeshPointsLabelZ = new QLabel(tr("Z Axis"), this);
  secondGridLayout->addWidget(blockMeshPointsLabelZ);
  blockMeshZMin = new QLineEdit(this);
  blockMeshZMin->setText("0");
  secondGridLayout->addWidget(blockMeshZMin);
  blockMeshZMax = new QLineEdit(this);
  blockMeshZMax->setText("1");
  secondGridLayout->addWidget(blockMeshZMax);

  QGridLayout* thirdGridLayout = new QGridLayout(this);
  thirdGridLayout->setMargin(pqPropertiesPanel::suggestedMargin());
  thirdGridLayout->setHorizontalSpacing(pqPropertiesPanel::suggestedHorizontalSpacing());
  thirdGridLayout->setVerticalSpacing(pqPropertiesPanel::suggestedVerticalSpacing());
  thirdGridLayout->setColumnStretch(0, 0);
  masterLayout->addLayout(thirdGridLayout, 0);

  QLabel* boundaryTypesLabel = new QLabel(tr("Select Background Mesh Boundary Types"),this);
  thirdGridLayout->addWidget(boundaryTypesLabel);
  blockBoundaryTable = new QTableWidget(this);
  blockBoundaryTable->setColumnCount(1);
  blockBoundaryTable->setRowCount(6);
  blockBoundaryTypes = new std::vector<QComboBox*>();
  QStringList blockTableLabels = { tr("Boundary Type") };
  QStringList boundaryTypes = { "Patch", "Wall", "Symmetry" };
  QStringList boundarySurfaces = { tr("Front"), tr("Back"), tr("Bottom"), tr("Top"), tr("Right"), tr("Left") };
  blockBoundaryTable->setHorizontalHeaderLabels(blockTableLabels);
  for (int i = 0; i<6; i++)
  {
    QComboBox *box = new QComboBox();
    blockBoundaryTypes->push_back(box);
    blockBoundaryTypes->at(i)->addItem(boundaryTypes[0]);
    blockBoundaryTypes->at(i)->addItem(boundaryTypes[1]);
    blockBoundaryTypes->at(i)->addItem(boundaryTypes[2]);
    blockBoundaryTable->setCellWidget(i, 0, blockBoundaryTypes->at(i));
  }
  blockBoundaryTable->setVerticalHeaderLabels(boundarySurfaces);
  thirdGridLayout->addWidget(blockBoundaryTable);

  QGridLayout* fourthGridLayout = new QGridLayout(this);
  fourthGridLayout->setMargin(pqPropertiesPanel::suggestedMargin());
  fourthGridLayout->setHorizontalSpacing(pqPropertiesPanel::suggestedHorizontalSpacing());
  fourthGridLayout->setVerticalSpacing(pqPropertiesPanel::suggestedVerticalSpacing());
  fourthGridLayout->setColumnStretch(1, 0);
  masterLayout->addLayout(fourthGridLayout, 0);

  //get cells in one axis
  QLabel* blockMeshXCellsLabel = new QLabel(tr("Cells Along X Direction"), this);
  fourthGridLayout->addWidget(blockMeshXCellsLabel);
  blockMeshXCells = new QLineEdit(this);
  blockMeshXCells->setText("8");
  fourthGridLayout->addWidget(blockMeshXCells);

  //SnappyHexMesh Parameters
  QLabel* snappyHexMeshLabel = new QLabel(tr("Snappy Hex Mesh Parameters"), this);
  fourthGridLayout->addWidget(snappyHexMeshLabel);
  fourthGridLayout->addWidget(spacer);

  QLabel* lineLevelLabel = new QLabel(tr("Line Level (All)"), this);
  fourthGridLayout->addWidget(lineLevelLabel);
  lineLevelValue = new QLineEdit(this);
  //lineLevelValue->setText("");
  fourthGridLayout->addWidget(lineLevelValue);

  QLabel* surfaceLevelLabel = new QLabel(tr("Surface Levels (All)"), this);
  fourthGridLayout->addWidget(surfaceLevelLabel);
  surfaceLevelValue = new QLineEdit(this);
  //surfaceLevelValue->setText("");
  fourthGridLayout->addWidget(surfaceLevelValue);

  QLabel* prismLevelLabel = new QLabel(tr("Prism Level"), this);
  fourthGridLayout->addWidget(prismLevelLabel);
  prismLevelValue = new QLineEdit(this);
  //prismLevelValue->setText("");
  fourthGridLayout->addWidget(prismLevelValue);

  //toggle advanced options
  advancedSnappyEnabled = new bool;
  *advancedSnappyEnabled = false;
  QPushButton* advancedSnappyParametersButton = new QPushButton(this);
  advancedSnappyParametersButton->setText(tr("Toggle Advanced SnappyHexMesh Options"));
  connect(advancedSnappyParametersButton, SIGNAL(clicked()), SLOT(toggleAdvancedSnappy()));
  fourthGridLayout->addWidget(advancedSnappyParametersButton);
  fourthGridLayout->addWidget(spacer);

  QGridLayout* fifthGridLayout = new QGridLayout(this);
  fifthGridLayout->setMargin(pqPropertiesPanel::suggestedMargin());
  fifthGridLayout->setHorizontalSpacing(pqPropertiesPanel::suggestedHorizontalSpacing());
  fifthGridLayout->setVerticalSpacing(pqPropertiesPanel::suggestedVerticalSpacing());
  fifthGridLayout->setColumnStretch(0, 0);
  masterLayout->addLayout(fifthGridLayout, 0);

  QLabel* advancedLevelsLabel = new QLabel(tr("Prescribe Levels for Each Tri-Surface file"));
  fifthGridLayout->addWidget(advancedLevelsLabel);
  advancedLevelValues = new QTableWidget(this);
  advancedLevelValues->setColumnCount(2);
  QStringList levelLabels = { "Line Level", "Surface Level"};
  advancedLevelValues->setHorizontalHeaderLabels(levelLabels);
  fifthGridLayout->addWidget(advancedLevelValues);

  QGridLayout* sixthGridLayout = new QGridLayout(this);
  sixthGridLayout->setMargin(pqPropertiesPanel::suggestedMargin());
  sixthGridLayout->setHorizontalSpacing(pqPropertiesPanel::suggestedHorizontalSpacing());
  sixthGridLayout->setVerticalSpacing(pqPropertiesPanel::suggestedVerticalSpacing());
  sixthGridLayout->setColumnStretch(1, 0);
  masterLayout->addLayout(sixthGridLayout, 0);

  QLabel* meshLocationLabel = new QLabel(tr("Mesh Location"), this);
  sixthGridLayout->addWidget(meshLocationLabel);
  meshLocationValue = new QComboBox(this);
  meshLocationValue->addItem(tr("External"));
  meshLocationValue->addItem(tr("Internal"));
  sixthGridLayout->addWidget(meshLocationValue);

  QLabel* cellsBetweenLevelsLabel = new QLabel(tr("Cells Between Levels"), this);
  sixthGridLayout->addWidget(cellsBetweenLevelsLabel);
  cellsBetweenLevelsValue = new QLineEdit(this);
  cellsBetweenLevelsValue->setText("3");
  sixthGridLayout->addWidget(cellsBetweenLevelsValue);

  QLabel* finalLayerThicknessLabel = new QLabel(tr("Final Layer Thickness"), this);
  sixthGridLayout->addWidget(finalLayerThicknessLabel);
  finalLayerThicknessValue = new QLineEdit(this);
  finalLayerThicknessValue->setText("0.5");
  sixthGridLayout->addWidget(finalLayerThicknessValue);

  QLabel* maxNonOrthoLabel = new QLabel(tr("Max Non-Orthogonality Angle"), this);
  sixthGridLayout->addWidget(maxNonOrthoLabel);
  maxNonOrthoValue = new QLineEdit(this);
  maxNonOrthoValue->setText("60");
  sixthGridLayout->addWidget(maxNonOrthoValue);

  QLabel* maxSkewnessLabel = new QLabel(tr("Max Skewness"), this);
  sixthGridLayout->addWidget(maxSkewnessLabel);
  maxSkewnessValue = new QLineEdit(this);
  maxSkewnessValue->setText("4");
  sixthGridLayout->addWidget(maxSkewnessValue);

  //generate mesh button
  QPushButton* generateMeshButton = new QPushButton(this);
  generateMeshButton->setText(tr("Generate Mesh"));
  connect(generateMeshButton, SIGNAL(clicked()), SLOT(generateMesh()));
  sixthGridLayout->addWidget(generateMeshButton);

  setLineEdits();

  //vars
  goodDocker = new bool();
  meshGenComplete = new bool();
  binary = new bool;
  *goodDocker = false;
  *binary = false;

  meshLocation = new std::vector<float>();
  advancedLineLevels = new std::vector<QLineEdit*>();
  advancedSurfaceLevels = new std::vector<QLineEdit*>();

  flowIterationsValue = 1000;
  
  preProcessFile = "preProcessing_snappyHexMesh.sh";
  meshGenLog = "logMeshGeneration.txt";
  dafoamVersion = "2.2.9";
  
  caseFileDir = new QString();
}

pqOptUMesh::~pqOptUMesh()
{
}

//-----------------------------------------------------------------------------

// about pvOptUMesh message
void pqOptUMesh::aboutMessage()
{
  QMessageBox about;
  about.setStyleSheet("QLabel{width: 400px; min-width: 600px; max-width: 400px;}");
  about.setText("<b><p align='center'><big>pvOptUMesh<br></b></p></big>");
  about.setStandardButtons(QMessageBox::Ok);
  about.setInformativeText("<p align='center'>A Paraview Mesh Generation GUI Plugin<br>Version: 0.1.0<br><br>© 2021 iDesign Lab, Iowa State University. All rights reserved.<br><br>Developers: Andrew Thomas and Ping He</p>");
  int ret = about.exec();
}

void pqOptUMesh::caseFileHandler()
{
  paths casePaths = pvOptFuncs::linkCaseFile(this);
  *caseFileDir = casePaths.caseFileDir;
  rootPath = casePaths.rootPath;
  dockerMessage();
}

// check if docker is good
void pqOptUMesh::dockerMessage()
{
  dockerFound info = pvOptFuncs::checkDocker(rootPath);
  *goodDocker = info.goodDocker;

  QMessageBox dockerInfo;
  dockerInfo.setText(info.success ? (*goodDocker ? tr("       Docker Successfully Found       ") : tr("      Docker Not Found     ")) : tr("       Error -- No Docker Info       "));
  dockerInfo.setStandardButtons(QMessageBox::Ok);
  dockerInfo.setInformativeText(*goodDocker ? tr(info.infoStr.c_str()) : tr("Without Docker, executing mesh generation script is impossible"));
  int ret = dockerInfo.exec();
}

void pqOptUMesh::selectSurface()
{
  if (caseFileDir->size()<1)
  {
    caseFileHandler();
  }

  QFileDialog dialog(this, tr("Select Tri-Surface Mesh Files"));
  dialog.setDirectory(rootPath.c_str());
  dialog.setFileMode(QFileDialog::ExistingFiles);

  if (dialog.exec())
  {
	  surfacePaths = dialog.selectedFiles();
  }
  else
  {
    return; //error
  }

  
  advancedLevelValues->setRowCount(surfacePaths.size());

  std::string constPath = rootPath + "/constant";
  triSurfacePath = constPath + "/triSurface";

  bool createdConstantDirectory = vtksys::SystemTools::MakeDirectory(triSurfacePath.c_str());
  
  QStringList fileLabels;
  int binaryCount = 0;
  for (int i = 0; i<surfacePaths.size(); i++)
  {
    QFile currentFile = surfacePaths[i];
    QFileInfo fileInfo(currentFile.fileName());
    std::string newPath = triSurfacePath + "/" + pvOptFuncs::removeExtensionFromFile(fileInfo.fileName().toStdString());
    /*if (isSTL(surfacePaths[i]))
    {*/
      //stlPath = newPath + ".stl";
      //currentFile.copy(stlPathc_str());
      if (isBinarySTL(surfacePaths[i].toStdString().c_str()))
      {
        binaryCount++;
        newPath += ".stlb";
      }
      else //is Ascii stl
      {
        newPath += ".stl";
      }
    /*}
    else
    {
      return; //error, other file types not yet supported
    }*/ 
    currentFile.copy(newPath.c_str());
    surfacePaths[i] = QString(newPath.c_str());
    
    QLineEdit *lineLevel = new QLineEdit();
    QLineEdit *surfaceLevel = new QLineEdit();
    advancedLineLevels->push_back(lineLevel);
    advancedSurfaceLevels->push_back(surfaceLevel);
    fileLabels << fileInfo.fileName().toStdString().c_str();
    advancedLevelValues->setCellWidget(i, 0, advancedLineLevels->at(i));
    advancedLevelValues->setCellWidget(i, 1, advancedSurfaceLevels->at(i));

  }

  if (binaryCount == surfacePaths.size())
  {
    changeLimits();
    *binary = true;
  }
  else
  {
    qDebug() << "non-binary";
    return; // give error
  }

  advancedLevelValues->setVerticalHeaderLabels(fileLabels);

}

bool pqOptUMesh::checkBlock()
{
  if ((blockMeshXMin->text().toDouble() >= blockMeshXMax->text().toDouble()) ||
      (blockMeshYMin->text().toDouble() >= blockMeshYMax->text().toDouble()) ||
      (blockMeshZMin->text().toDouble() >= blockMeshZMax->text().toDouble())) 
    {
      QMessageBox blockMeshWarning;
      blockMeshWarning.setText(tr("One or more of the block points are invalid, check that all maxima are greater than minima"));
      blockMeshWarning.setStandardButtons(QMessageBox::Ok);
      int ret = blockMeshWarning.exec();
      return false;
    }
  return true;
}

bool pqOptUMesh::checkLocation() //fix
{
  //print warning if location is not inside block
  if ((meshLocation->at(0) < blockMeshXMin->text().toDouble() || meshLocation->at(0) > blockMeshXMax->text().toDouble()) ||
      (meshLocation->at(1) < blockMeshYMin->text().toDouble() || meshLocation->at(1) > blockMeshYMax->text().toDouble()) ||
      (meshLocation->at(2) < blockMeshZMin->text().toDouble() || meshLocation->at(2) > blockMeshZMax->text().toDouble()))
  {
    QMessageBox locationWarning;
    std::string warning = "Please expand the block to include the point (" + std::to_string(meshLocation->at(0)) + 
                          ", " + std::to_string(meshLocation->at(1)) + ", " + std::to_string(meshLocation->at(2)) + ")";
    locationWarning.setText(tr(warning.c_str()));
    locationWarning.setStandardButtons(QMessageBox::Ok);
    int ret = locationWarning.exec();
    return false;
  }
  return true; //location is inside block
}

/* bool pqOptUMesh::checkUnset()
{

}*/

void pqOptUMesh::toggleAdvancedSnappy()
{
  if (*advancedSnappyEnabled)
  {
    *advancedSnappyEnabled = false;
  }
  else
  {
    *advancedSnappyEnabled = true;
  }
  setLineEdits();
}

void pqOptUMesh::setLineEdits()
{
  meshLocationValue->setEnabled(*advancedSnappyEnabled);
  cellsBetweenLevelsValue->setEnabled(*advancedSnappyEnabled);
  finalLayerThicknessValue->setEnabled(*advancedSnappyEnabled);
  maxNonOrthoValue->setEnabled(*advancedSnappyEnabled);
  maxSkewnessValue->setEnabled(*advancedSnappyEnabled);
  advancedLevelValues->setEnabled(*advancedSnappyEnabled);

  lineLevelValue->setEnabled(!*advancedSnappyEnabled);
  surfaceLevelValue->setEnabled(!*advancedSnappyEnabled);
}

void pqOptUMesh::changeLimits() //round vals
{
  std::vector<float> totalLimits;
  for (int i = 0; i<surfacePaths.size(); i++)
  {
    std::vector<std::vector<vertex>> triangles = STLImport(surfacePaths[i].toStdString().c_str());
    std::vector<float> limits = findBounds(triangles); //xMin xMax yMin yMax zMin zMax
    
    if (i == 0)
    {
      totalLimits = limits;
    }
    else
    {
      if (limits[0] < totalLimits[0])
      {
        totalLimits[0] = limits[0];
      }
      else if (limits[1] > totalLimits[1])
      {
        totalLimits[1] = limits[1];
      }

      if (limits[2] < totalLimits[2])
      {
        totalLimits[2] = limits[2];
      }
      else if (limits[3] > totalLimits[3])
      {
        totalLimits[3] = limits[3];
      }

      if (limits[4] < totalLimits[4])
      {
        totalLimits[4] = limits[4];
      }
      else if (limits[5] > totalLimits[5])
      {
        totalLimits[5] = limits[5];
      }
    }
  } 

  blockMeshXMin->setText(tr(std::to_string(totalLimits[0]).c_str()));
  blockMeshXMax->setText(tr(std::to_string(totalLimits[1]).c_str()));
  blockMeshYMin->setText(tr(std::to_string(totalLimits[2]).c_str()));
  blockMeshYMax->setText(tr(std::to_string(totalLimits[3]).c_str()));
  blockMeshZMin->setText(tr(std::to_string(totalLimits[4]).c_str()));
  blockMeshZMax->setText(tr(std::to_string(totalLimits[5]).c_str()));
}

void pqOptUMesh::getLocation()
{
  std::vector<float> loc{ 0.0, 0.0, 0.0 };
  if (binary) //change to check if all stl
  {
    for (int i = 0; i<surfacePaths.size(); i++)
    {
      std::vector<std::vector<vertex>> triangles = STLImport(surfacePaths[i].toStdString().c_str());
  
      std::vector<float> location = findLocation(triangles, meshLocationValue->currentIndex());

      if ((meshLocationValue->currentIndex() == 0) && (pvOptFuncs::vectorNorm(location) > pvOptFuncs::vectorNorm(loc)))
      {
        loc = location;
      }
      else if (meshLocationValue->currentIndex() == 1)
      {
        *meshLocation = location;
        return;
      }
    }
    *meshLocation = loc;
  }
}

void pqOptUMesh::visualizeBoundBox() //vtkSTLReader not found?? CMake??
{
  // Create the RenderWindow, Renderer and both Actors
  //
  vtkSmartPointer<vtkRenderer> ren1 = vtkSmartPointer<vtkRenderer>::New();
  vtkSmartPointer<vtkRenderWindow> renWin = vtkSmartPointer<vtkRenderWindow>::New();
  renWin->AddRenderer(ren1);
  renWin->SetWindowName("PvOptUMesh");

  vtkSmartPointer<vtkRenderWindowInteractor> iren =
    vtkSmartPointer<vtkRenderWindowInteractor>::New();
  iren->SetRenderWindow(renWin);

  vtkNew<vtkNamedColors> namedColors;

  // Create Pipeline
  vtkSmartPointer<vtkCubeSource> box = vtkSmartPointer<vtkCubeSource>::New();
  box->SetBounds(blockMeshXMin->text().toDouble(), blockMeshXMax->text().toDouble(), 
                  blockMeshYMin->text().toDouble(), blockMeshYMax->text().toDouble(), 
                  blockMeshZMin->text().toDouble(), blockMeshZMax->text().toDouble());
  
  vtkSmartPointer<vtkAssembly> assembly = vtkSmartPointer<vtkAssembly>::New();
  for (int i = 0; i < surfacePaths.size(); i++)
  {
    /*vtkSmartPointer<vtkPolyDataMapper> stlMapper = vtkSmartPointer<vtkPolyDataMapper>::New();
    vtkSmartPointer<vtkPolyData> stlPolyData = vtkSmartPointer<vtkPolyData>::New();
    stlMapper->SetInputData(stlPolyData);*/
    
    vtkNew<vtkSTLReader> reader;
    reader->SetFileName(surfacePaths[i].toStdString().c_str());
    reader->Update();
    vtkSmartPointer<vtkPolyDataMapper> stlMapper = vtkSmartPointer<vtkPolyDataMapper>::New();
    stlMapper->SetInputConnection(reader->GetOutputPort());

    vtkSmartPointer<vtkActor> stlActor = vtkSmartPointer<vtkActor>::New();
    stlActor->SetMapper(stlMapper);
    //stlActor->GetProperty()->SetDiffuseColor(namedColors->GetColor3d("LightSteelBlue").GetData());
    assembly->AddPart(stlActor);
  }
  
  //vtkSmartPointer<vtkPolyDataMapper> mapper = vtkSmartPointer<vtkPolyDataMapper>::New();
  vtkSmartPointer<vtkFeatureEdges> edges = vtkSmartPointer<vtkFeatureEdges>::New();
  edges->SetInputConnection(box->GetOutputPort());
  vtkSmartPointer<vtkPolyDataMapper> mapper = vtkSmartPointer<vtkPolyDataMapper>::New();
  mapper->SetInputConnection(edges->GetOutputPort());
  vtkSmartPointer<vtkActor> boxActor = vtkSmartPointer<vtkActor>::New();
  boxActor->SetMapper(mapper);
  //boxActor->GetProperty()->SetColor(namedColors->GetColor3d("Tomato").GetData());

  // Create the widget and its representation
  vtkSmartPointer<vtkBorderRepresentation> rep = vtkSmartPointer<vtkBorderRepresentation>::New();
  rep->ProportionalResizeOn();
  rep->SetShowBorderToOn();

  vtkSmartPointer<vtkBorderWidget> widget = vtkSmartPointer<vtkBorderWidget>::New();
  widget->SetInteractor(iren);
  widget->SetRepresentation(rep);
  widget->SelectableOff();

  // Add the actors to the renderer, set the background and size
  //
  ren1->AddActor(boxActor);
  ren1->AddActor(assembly);
  ren1->SetBackground(0.1, 0.2, 0.4);
  renWin->SetSize(600, 600);

  // record events
  /*vtkSmartPointer<vtkInteractorEventRecorder> recorder =
    vtkSmartPointer<vtkInteractorEventRecorder>::New();
  recorder->SetInteractor(iren);

  #ifdef RECORD
    recorder->SetFileName("record.log");
    recorder->On();
    recorder->Record();
  #else
    recorder->ReadFromInputStringOn();
    recorder->SetInputString(BorderWidgetEventLog);
  #endif*/
  // render the image
  //
  iren->Initialize();
  renWin->Render();
  widget->On();

  // Remove the observers so we can go interactive. Without this the "-I"
  // testing option fails.

  /*#ifndef RECORD
    recorder->Play();
    recorder->Off();
  #endif*/

  iren->Start();
}

/*void pqOptUMesh::visualizeBoundBoxPy()
{
  //write python script
  std::ostringstream vizPath;
  vizPath << rootPath << "/visualizeBackgroundMesh.py";
  std::ofstream vizBlockFile(vizPath.str().c_str(), std::ios::out);

  vizBlockFile <<
  "import paraview.simple as psimple\n"

}*/

std::string pqOptUMesh::writePreProcessing()
{
  std::ostringstream preProcessPath;
  preProcessPath << rootPath << "/" << preProcessFile;
  #ifdef _WIN32
    std::ofstream preProcessing(preProcessPath.str().c_str(), std::ios::binary);
  #else
    std::ofstream preProcessing(preProcessPath.str().c_str(), std::ios::out);
  #endif
  
  preProcessing<<
  "#!/bin/bash\n"
  "\n"
  "if [ -z \"$WM_PROJECT\" ]; then\n"
  "  echo \"OpenFOAM environment not found, forgot to source the OpenFOAM bashrc?\"\n"
  "  exit\n"
  "fi\n"
  "\n"
  "# pre-processing\n"
  "\n"
  "# generate mesh\n"
  "echo \"Generating mesh..\"\n"
  "\n"
  "echo \"Running snappyHexMesh mesh..\"\n"
  "blockMesh &> "<<meshGenLog<<"\n"
  "surfaceFeatureExtract >> "<<meshGenLog<<"\n"
  "snappyHexMesh -overwrite >> "<<meshGenLog<<"\n"
  //"createPatch -overwrite >> "<<meshGenLog<<"\n"
  "renumberMesh -overwrite >> "<<meshGenLog<<"\n"
  "echo \"Generating mesh.. Done\"" << std::endl;

  preProcessing.flush();
  preProcessing.close();

  return preProcessPath.str();
}

void pqOptUMesh::controlDict(std::string systemPath)
{
  std::ostringstream controlDictPath;
  controlDictPath << systemPath << "/controlDict";
  std::ofstream controlDictFile(controlDictPath.str().c_str(), std::ios::out);

  controlDictFile << 
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       dictionary;\n"
  "    location    \"system\";\n"
  "    object      controlDict;\n"
  "}\n"
  "\n"
  "startFrom       startTime;\n"
  "startTime       0;\n"
  "stopAt          endTime;\n"
  "endTime         "<<flowIterationsValue<<";\n"
  "deltaT          1;\n"
  "writeControl    timeStep;\n"
  "writeInterval   "<<flowIterationsValue<<";\n"
  "purgeWrite      0;\n"
  "writeFormat     ascii;\n"
  "writePrecision  16;\n"
  "writeCompression on;\n"
  "timeFormat      general;\n"
  "timePrecision   16;\n"
  "runTimeModifiable true;\n"
  "DebugSwitches\n"
  "{\n"
  "    SolverPerformance 0;\n"
  "}\n";

  controlDictFile.flush();
  controlDictFile.close();
}

void pqOptUMesh::fvSchemes(std::string systemPath)
{
  std::ostringstream fvSchemesPath;
  fvSchemesPath << systemPath << "/fvSchemes";
  std::ofstream fvSchemesFile(fvSchemesPath.str().c_str(), std::ios::out);

  fvSchemesFile << 
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       dictionary;\n"
  "    location    \"system\";\n"
  "    object      fvSchemes;\n"
  "}\n"
  "\n"
  "ddtSchemes\n" 
  "{\n"
  "    default                                             steadyState;\n"
  "}\n"
  "\n"
  "gradSchemes\n"
  "{\n"
  "    default                                             Gauss linear;\n"
  "}\n"
  "\n"
  "divSchemes\n"
  "{\n"
  "    default                                             none;\n"
  "}\n"
  "\n"
  "interpolationSchemes\n"
  "{\n"
  "    default                                             linear;\n"
  "}\n"
  "\n"
  "laplacianSchemes\n"
  "{\n"
  "    default                                             Gauss linear corrected;\n"
  "}\n"
  "\n"
  "snGradSchemes\n"
  "{\n"
  "    default                                             corrected;\n"
  "}\n"
  "\n"
  "fluxRequired\n"
  "{\n"
  "    default                                             no;\n"
  "}\n"
  "\n"
  "wallDist\n"
  "{\n"
  "    method                                              meshWave;\n"
  "}\n";

  fvSchemesFile.flush();
  fvSchemesFile.close();
}

void pqOptUMesh::fvSolution(std::string systemPath)
{
  std::ostringstream fvSolutionPath;
  fvSolutionPath << systemPath << "/fvSolution";
  std::ofstream fvSolutionFile(fvSolutionPath.str().c_str(), std::ios::out);

  fvSolutionFile << 
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       dictionary;\n"
  "    location    \"system\";\n"
  "    object      fvSolution;\n"
  "}\n"
  "\n"
  "SIMPLE\n"
  "{\n"
  "    nNonOrthogonalCorrectors           0;\n"
  "}\n";

  fvSolutionFile.flush();
  fvSolutionFile.close();
}

void pqOptUMesh::blockMeshDict(std::string systemPath)
{
  int xCells = blockMeshXCells->text().toInt();
  double cellDensity = xCells / (blockMeshXMax->text().toDouble() - blockMeshXMin->text().toDouble());
  int yCells = int(cellDensity * (blockMeshYMax->text().toDouble() - blockMeshYMin->text().toDouble()));
  int zCells = int(cellDensity * (blockMeshZMax->text().toDouble() - blockMeshZMin->text().toDouble()));

  std::ostringstream blockMeshDictPath;
  blockMeshDictPath << systemPath << "/blockMeshDict";
  std::ofstream blockMeshFile(blockMeshDictPath.str().c_str(), std::ios::out);

  blockMeshFile <<
  "FoamFile\n"
  "{\n"
  "    version 2.0;\n"
  "    format ascii;\n"
  "    class dictionary;\n"
  "    location system;\n"
  "    object blockMeshDict;\n"
  "}\n"
  "\n"
  "convertToMeters 1;\n"
  "\n"
  "vertices\n"
  "(\n"
  "    ("<<blockMeshXMin->text().toDouble()<<" "<<blockMeshYMin->text().toDouble()<<" "<<blockMeshZMin->text().toDouble()<<")\n"
  "    ("<<blockMeshXMax->text().toDouble()<<" "<<blockMeshYMin->text().toDouble()<<" "<<blockMeshZMin->text().toDouble()<<")\n"
  "    ("<<blockMeshXMax->text().toDouble()<<" "<<blockMeshYMax->text().toDouble()<<" "<<blockMeshZMin->text().toDouble()<<")\n"
  "    ("<<blockMeshXMin->text().toDouble()<<" "<<blockMeshYMax->text().toDouble()<<" "<<blockMeshZMin->text().toDouble()<<")\n"
  "    ("<<blockMeshXMin->text().toDouble()<<" "<<blockMeshYMin->text().toDouble()<<" "<<blockMeshZMax->text().toDouble()<<")\n"
  "    ("<<blockMeshXMax->text().toDouble()<<" "<<blockMeshYMin->text().toDouble()<<" "<<blockMeshZMax->text().toDouble()<<")\n"
  "    ("<<blockMeshXMax->text().toDouble()<<" "<<blockMeshYMax->text().toDouble()<<" "<<blockMeshZMax->text().toDouble()<<")\n"
  "    ("<<blockMeshXMin->text().toDouble()<<" "<<blockMeshYMax->text().toDouble()<<" "<<blockMeshZMax->text().toDouble()<<")\n"
  ");\n"
  "\n"
  "blocks  \n"
  "(\n"
  "    hex ( 0 1 2 3 4 5 6 7)\n"
  "    ("<<xCells<<" "<<yCells<<" "<<zCells<<")\n"
  "    simpleGrading (1 1 1)\n"
  ");\n"
  "\n"
  "edges  \n"
  "(\n"
  ");\n"
  "\n"
  "patches  \n"
  "( \n"
  "  "<<blockBoundaryTypes->at(0)->currentText().toStdString()<<" front\n"
  "  (\n"
  "      (0 4 7 3)\n"
  "  ) \n"
  "  "<<blockBoundaryTypes->at(1)->currentText().toStdString()<<" back\n"
  "  (\n"
  "      (1 2 6 5)\n"
  "  ) \n"
  "  "<<blockBoundaryTypes->at(2)->currentText().toStdString()<<" bot\n"
  "  (\n"
  "      (0 1 5 4)\n"
  "  ) \n"
  "  "<<blockBoundaryTypes->at(3)->currentText().toStdString()<<" top\n"
  "  (\n"
  "      (3 7 6 2)\n"
  "  ) \n"
  "  "<<blockBoundaryTypes->at(4)->currentText().toStdString()<<" right\n"
  "  (\n"
  "      (0 3 2 1)\n"
  "  ) \n"
  "  "<<blockBoundaryTypes->at(5)->currentText().toStdString()<<" left\n"
  "  (\n"
  "      (4 5 6 7)\n"
  "  )\n"
  ");\n"
  "\n"
  "mergePatchPairs  \n"
  "(\n"
  ");" << std::endl;

  blockMeshFile.flush();
  blockMeshFile.close();
}

void pqOptUMesh::surfaceFeatureExtractDict(std::string systemPath)
{
  std::ostringstream surfaceFeatureExtractDictPath;
  surfaceFeatureExtractDictPath << systemPath << "/surfaceFeatureExtractDict";
  std::ofstream surfaceFeatureExtractFile(surfaceFeatureExtractDictPath.str().c_str(), std::ios::out);

  surfaceFeatureExtractFile << 
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       dictionary;\n"
  "    object      surfaceFeatureExtractDict;\n"
  "}\n" << std::endl;

  for (int i = 0; i<surfacePaths.size(); i++)
  {
    surfaceFeatureExtractFile <<
    surfacePaths[i].toStdString().c_str() << "\n"
    "{\n"
    "    extractionMethod    extractFromSurface;\n"
    "    includedAngle       150;\n"
    "    geometricTestOnly   yes;\n"
    "    intersectionMethod  none;\n"
    "    writeObj            yes;\n"
    "}\n" << std::endl;
  }

  surfaceFeatureExtractFile.flush();
  surfaceFeatureExtractFile.close();
}

void pqOptUMesh::snappyHexMeshDict(std::string systemPath)
{
  std::vector<int> lineLevels;
  std::vector<double> surfaceLevels;
  int prismLayer, cellsBetweenLevels, maxNonOrtho, maxSkewness;
  double finalLayerThickness;
  std::string meshName;

  prismLayer = prismLevelValue->text().toInt();
  cellsBetweenLevels = cellsBetweenLevelsValue->text().toInt();
  finalLayerThickness = finalLayerThicknessValue->text().toDouble();
  maxNonOrtho = maxNonOrthoValue->text().toInt();
  maxSkewness = maxSkewnessValue->text().toInt();

  getLocation();
  
  for (int i = 0; i<surfacePaths.size(); i++)
  {
    lineLevels.push_back(advancedLineLevels->at(i)->text().toInt());
    surfaceLevels.push_back(advancedSurfaceLevels->at(i)->text().toDouble());
  }

  std::ostringstream snappyHexMeshDictPath;
  snappyHexMeshDictPath << systemPath << "/snappyHexMeshDict";
  std::ofstream snappyHexMeshFile(snappyHexMeshDictPath.str().c_str(), std::ios::out);

  snappyHexMeshFile <<
  "FoamFile\n"
  "{\n"
  "    version 2.0;\n"
  "    format ascii;\n"
  "    class dictionary;\n"
  "    location \"system\";\n"
  "    object snappyHexMeshDict;\n"
  "}\n"
  "\n"
  "    prismLayer "<<prismLayer<<";\n"
  "\n"
  "    castellatedMesh true;\n"
  "    snap true;\n"
  "    addLayers true;\n"
  "    geometry\n"
  "    {\n" << std::endl;

  for (int i = 0; i<surfacePaths.size(); i++)
  {
    QFile currentFile = surfacePaths[i];
    QFileInfo fileInfo(currentFile.fileName());
    meshName = pvOptFuncs::removeExtensionFromFile(fileInfo.fileName().toStdString());

    snappyHexMeshFile <<
    "        "<<fileInfo.fileName().toStdString()<<"\n"
    "        {\n"
    "            type triSurfaceMesh;\n"
    "            name  "<<meshName<<";\n"
    "        }\n"
    "\n" << std::endl;
  }
  
  snappyHexMeshFile <<
  "    }\n"
  "\n"
  "    castellatedMeshControls\n"
  "    {\n"
  "        features\n"
  "        (\n" << std::endl;

  for (int i = 0; i<surfacePaths.size(); i++)
  {
    QFile currentFile = surfacePaths[i];
    QFileInfo fileInfo(currentFile.fileName());
    meshName = pvOptFuncs::removeExtensionFromFile(fileInfo.fileName().toStdString());
    
    if (*advancedSnappyEnabled)
    {
      snappyHexMeshFile <<
      "            {\n"
      "                file \""<<meshName<<".eMesh\";\n"
      "                level "<<lineLevels[i]<<";\n"
      "            }\n"
      "\n" << std::endl;
    }
    else
    {
      snappyHexMeshFile <<
      "            {\n"
      "                file \""<<meshName<<".eMesh\";\n"
      "                level "<<lineLevelValue->text().toInt()<<";\n"
      "            }\n"
      "\n" << std::endl;
    }
  }

  snappyHexMeshFile <<
  "        );\n"
  "        refinementSurfaces\n"
  "        {\n" << std::endl;

  for (int i = 0; i<surfacePaths.size(); i++)
  {
    QFile currentFile = surfacePaths[i];
    QFileInfo fileInfo(currentFile.fileName());
    meshName = pvOptFuncs::removeExtensionFromFile(fileInfo.fileName().toStdString());

    if (*advancedSnappyEnabled)
    {
      fraction frac = pvOptFuncs::decimalToFraction(surfaceLevels[i]);
      if (frac.b == 1) 
      {
        frac.b = frac.a;
      }

      snappyHexMeshFile <<
      "            "<<meshName<<"\n"
      "            {\n"
      "                level ( "<<frac.a<<" "<<frac.b<<" );\n"
      "            }\n"
      "\n" << std::endl;
    }
    else
    {
      fraction frac = pvOptFuncs::decimalToFraction(surfaceLevelValue->text().toDouble());
      if (frac.b == 1) 
      {
        frac.b = frac.a;
      }

      snappyHexMeshFile <<
      "            "<<meshName<<"\n"
      "            {\n"
      "                level ( "<<frac.a<<" "<<frac.b<<" );\n"
      "            }\n"
      "\n" << std::endl;
    }
  }

  snappyHexMeshFile <<
  "        }\n"
  "\n"
  "        refinementRegions\n"
  "        {\n"
  "\n"
  "           \n"
  "        }\n"
  "\n"
  "        locationInMesh ( "<<meshLocation->at(0)<<" "<<meshLocation->at(1)<<" "<<meshLocation->at(2)<<" );\n"
  "        maxLocalCells 10000000;\n"
  "        maxGlobalCells 200000000;\n"
  "        minRefinementCells 0;\n"
  "        nCellsBetweenLevels "<<cellsBetweenLevels<<";\n"
  "        resolveFeatureAngle 10;\n"
  "        allowFreeStandingZoneFaces true;\n"
  "        planarAngle 30;\n"
  "        maxLoadUnbalance 0.10;\n"
  "    }\n"
  "\n"
  "    snapControls\n"
  "    {\n"
  "        nSolveIter 50;\n"
  "        nSmoothPatch 3;\n"
  "        tolerance 2.0;\n"
  "        nRelaxIter 5;\n"
  "        nFeatureSnapIter 10;\n"
  "        implicitFeatureSnap false;\n"
  "        explicitFeatureSnap true;\n"
  "        multiRegionFeatureSnap false;\n"
  "    }\n"
  "\n"
  "    addLayersControls\n"
  "    {\n"
  "        layers\n"
  "        {\n" << std::endl;

  //Correct?
  for (int i = 0; i<surfacePaths.size(); i++)
  {
    QFile currentFile = surfacePaths[i];
    QFileInfo fileInfo(currentFile.fileName());
    meshName = pvOptFuncs::removeExtensionFromFile(fileInfo.fileName().toStdString());

    snappyHexMeshFile <<
    "            \""<<meshName<<".*\"\n"
    "            {\n"
    "              nSurfaceLayers $prismLayer;\n"
    "            }\n"
    "\n" << std::endl;
  }

  snappyHexMeshFile <<
  "        }\n"
  "\n"
  "        relativeSizes true;\n"
  "        expansionRatio 1.3;\n"
  "        finalLayerThickness "<<finalLayerThickness<<";\n"
  "        minThickness 0.001;\n"
  "        nGrow 0;\n"
  "        featureAngle 180;\n"
  "        slipFeatureAngle 80;\n"
  "        nRelaxIter 5;\n"
  "        nSmoothSurfaceNormals 1;\n"
  "        nSmoothNormals 3;\n"
  "        nSmoothThickness 10;\n"
  "        maxFaceThicknessRatio 0.5;\n"
  "        maxThicknessToMedialRatio 0.3;\n"
  "        minMedialAxisAngle 90;\n"
  "        nBufferCellsNoExtrude 0;\n"
  "        nLayerIter 50;\n"
  "        nRelaxedIter 20;\n"
  "        writeVTK false;\n"
  "        noErrors false;\n"
  "        layerRecovery 1;\n"
  "        growZoneLayers false;\n"
  "        projectGrownUp 0.0;\n"
  "    }\n"
  "\n"
  "    meshQualityControls\n"
  "    {\n"
  "        maxNonOrtho "<<maxNonOrtho<<";\n"
  "        maxBoundarySkewness "<<maxSkewness<<";\n"
  "        maxInternalSkewness "<<maxSkewness<<";\n"
  "        maxConcave 80;\n"
  "        minFlatness 0.5;\n"
  "        minVol 1.00E-13;\n"
  "        minTetQuality -1e30;\n"
  "        minArea -1;\n"
  "        minTwist 0.03;\n"
  "        minDeterminant 0.01;\n"
  "        minFaceWeight 0.03;\n"
  "        minVolRatio 0.01;\n"
  "        minTriangleTwist -1;\n"
  "        nSmoothScale 4;\n"
  "        errorReduction 0.75;\n"
  "        relaxed {maxNonOrtho "<<maxNonOrtho<<"; maxBoundarySkewness "<<maxSkewness<<"; }\n"
  "\n"
  "    }\n"
  "\n"
  "    debug 0;\n"
  "    mergeTolerance 1E-6;\n"
  "    autoBlockMesh false;" << std::endl;
}

int pqOptUMesh::generateMesh()
{
  //Double checking conflicts/incomplete data
  if (caseFileDir->size()<1)
  {
    caseFileHandler();
  }
  if (surfacePaths.size()<1)
  {
    selectSurface();
  }
  if (!checkBlock())
  {
    qDebug() << "bad block";
    return -1;
  }
  /*if (!checkLocation())
  {
    qDebug() << "bad location";
    return -1;
  }*/
  /* if (!checkUnset())
  {
    qDebug() << "Parameters unset";
    return -1;
  }*/

  bool success = false;
  std::string detailedOutput;
  std::string preProcessingPath;

  std::string systemPath = rootPath + "/system";
  bool createdDirectory = vtksys::SystemTools::MakeDirectory(systemPath.c_str());
  if (!createdDirectory) 
  { 
    qDebug() << "Could not create system dir";
    return 1; 
  }
  controlDict(systemPath);
  fvSchemes(systemPath);
  fvSolution(systemPath);
  blockMeshDict(systemPath);
  surfaceFeatureExtractDict(systemPath);
  snappyHexMeshDict(systemPath);
  preProcessingPath = writePreProcessing();

  if (*goodDocker)
  {
    std::string cmdPreProcessing = "docker run -i --rm -u dafoamuser --mount \"type=bind,src=" + rootPath +
    ",target=/home/dafoamuser/mount\" -w /home/dafoamuser/mount dafoam/opt-packages:v" + dafoamVersion + " bash -c " +
    "\". /home/dafoamuser/dafoam/loadDAFoam.sh && ./" + preProcessFile + "\"";

    QFuture<void> futureMeshGen = QtConcurrent::run([=]() {
      pvOptFuncs::executeSystemCmd(cmdPreProcessing);
    });
    futureMeshGen.waitForFinished();

    //Gather output for message box
    std::string logPath = rootPath + "/" + meshGenLog;
    success = true;
    QFile file(logPath.c_str());
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
    {
      success = false;
    }
    while (!file.atEnd()) 
    {
      QByteArray line = file.readLine();
      detailedOutput.append(line.toStdString().c_str());
    }
    file.close();
  }

  QMessageBox meshGenInfo;
  meshGenInfo.setText(success ? tr("       Mesh Generation Complete       ") : (*goodDocker ? tr("       Error -- No Output File       ") : tr("Docker Not Installed")));
  meshGenInfo.setStandardButtons(QMessageBox::Ok);
  std::string infoMsg = "";
  if (*goodDocker)
  {
    meshGenInfo.setInformativeText(tr(infoMsg.c_str()));
    meshGenInfo.setDetailedText(tr(detailedOutput.c_str()));
  }
  int ret = meshGenInfo.exec();

  *meshGenComplete = true;
  return 0;
}